#!/usr/bin/env python
#
# cloup documentation build configuration file, created by
# sphinx-quickstart on Fri Jun  9 13:47:02 2017.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

# If extensions (or modules to document with autodoc) are in another
# directory, add these directories to sys.path here. If the directory is
# relative to the documentation root, use os.path.abspath to make it
# absolute, like shown here.
#
import os

import cloup

# import sys
# sys.path.insert(0, os.path.abspath('..'))
PROJ_DIR = os.path.abspath(os.path.join(__file__, '..', '..'))

# -- General configuration ---------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = [
    'autoapi.extension',
    'sphinx.ext.intersphinx',
    'sphinx.ext.viewcode',
    'sphinx.ext.autodoc.typehints',
    'sphinx_panels',
    'sphinx_copybutton',  # adds a copy button to code blocks
]

autoclass_content = 'both'
autodoc_typehints = 'description'

autoapi_type = 'python'
autoapi_dirs = [os.path.join(PROJ_DIR, 'cloup')]
autoapi_template_dir = '_autoapi_templates'
autoapi_keep_files = True
# autoapi_add_toctree_entry = False
autoapi_options = [
    'members',
    'undoc-members',
    'show-inheritance',
    'show-module-summary',
    'special-members',
    'imported-members'
]

intersphinx_mapping = {
    'python': ('https://docs.python.org/', None),
    'Click': ('https://click.palletsprojects.com', None)
}

# Add any paths that contain templates here, relative to this directory.
templates_path = [autoapi_template_dir]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = 'cloup'
copyright = "2020, Gianluca Gippetto"
author = "Gianluca Gippetto"

# The version info for the project you're documenting, acts as replacement
# for |version| and |release|, also used in various other places throughout
# the built documents.
#
# The short X.Y version.
version = cloup.__version__
# The full version, including alpha/beta/rc tags.
release = cloup.__version__

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "en"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'default'

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False

# -- Options for HTML output -------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_title = "Cloup v{version}".format(version=version)
html_theme = "sphinx_book_theme"
html_theme_options = {
    "repository_url": "https://github.com/janluke/cloup",
    "path_to_docs": "docs",
    "use_repository_button": True,
    "use_issues_button": True,
    "use_edit_page_button": True,
    "home_page_in_toc": False,
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

def primary_color_rgba(alpha=1.0):
    return "rgba(0, 113, 188, %f)" % alpha

panels_css_variables = {
    "tabs-color-label-active": primary_color_rgba(),
    "tabs-color-label-inactive": primary_color_rgba(0.5),
    "tabs-color-overline": primary_color_rgba(0.2),
    # "tabs-color-underline": "rgb(207, 236, 238)",
    # "tabs-size-label": "1rem",
}
panels_add_bootstrap_css = False
html_css_files = [
    'styles/copybutton-overrides.css',
    'styles/tabs-overrides.css',
    'styles/theme-overrides.css',
]

# -- Options for HTMLHelp output ---------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = 'cloupdoc'

# -- Options for LaTeX output ------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',

    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',

    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',

    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass
# [howto, manual, or own class]).
latex_documents = [
    (master_doc, 'cloup.tex',
     'cloup Documentation',
     'Gianluca Gippetto', 'manual'),
]

# -- Options for manual page output ------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (master_doc, 'cloup',
     'cloup Documentation',
     [author], 1)
]

# -- Options for Texinfo output ----------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (master_doc, 'cloup',
     'cloup Documentation',
     author,
     'cloup',
     'One line description of project.',
     'Miscellaneous'),
]



from pytest import fixture
from tests.example_command import make_example_command
from tests.example_group import make_example_group


@fixture(scope='session')
def get_example_command():
    aligned_cmd = make_example_command(align_option_groups=True)
    non_aligned_cmd = make_example_command(align_option_groups=False)

    def get_command(align_option_groups):
        return aligned_cmd if align_option_groups else non_aligned_cmd

    return get_command


@fixture(scope='session')
def get_example_group():
    aligned_cmd = make_example_group(align_sections=True)
    non_aligned_cmd = make_example_group(align_sections=False)

    def get_group(align_sections):
        return aligned_cmd if align_sections else non_aligned_cmd

    return get_group



# flake8: noqa E128
import click

import cloup
from cloup import option, option_group


def make_example_command(align_option_groups):
    @cloup.command(
        'clouptest',
        align_option_groups=align_option_groups,
        context_settings={'terminal_width': 80})
    @option_group(
        'Option group A',
        option('--one', help='1st option of group A'),
        option('--two', help='2nd option of group A'),
        option('--three', help='3rd option of group A'),
        help='This is a very useful description of group A')
    @option_group(
        'Option group B',
        option('--four / --no-four', help='1st option of group B'),
        option('--five', help='2nd option of group B', hidden=True),
        option('--six', help='3rd option of group B'))
    @option('--seven', help='first uncategorized option',
            type=click.Choice('yes no ask'.split()))
    @option('--height', help='second uncategorized option')
    @option('--nine', help='second uncategorized option', hidden=True)
    def cmd(**kwargs):
        """ A CLI that does nothing. """
        print(kwargs)

    cmd.expected_help = (_EXPECTED_ALIGNED_HELP if align_option_groups
                         else _EXPECTED_NON_ALIGNED_HELP)
    return cmd


_EXPECTED_ALIGNED_HELP = """
Usage: clouptest [OPTIONS]

  A CLI that does nothing.

Option group A:
  This is a very useful description of group A
  --one TEXT            1st option of group A
  --two TEXT            2nd option of group A
  --three TEXT          3rd option of group A

Option group B:
  --four / --no-four    1st option of group B
  --six TEXT            3rd option of group B

Other options:
  --seven [yes|no|ask]  first uncategorized option
  --height TEXT         second uncategorized option
  --help                Show this message and exit.
    """.strip()

_EXPECTED_NON_ALIGNED_HELP = """
Usage: clouptest [OPTIONS]

  A CLI that does nothing.

Option group A:
  This is a very useful description of group A
  --one TEXT    1st option of group A
  --two TEXT    2nd option of group A
  --three TEXT  3rd option of group A

Option group B:
  --four / --no-four  1st option of group B
  --six TEXT          3rd option of group B

Other options:
  --seven [yes|no|ask]  first uncategorized option
  --height TEXT         second uncategorized option
  --help                Show this message and exit.
""".strip()

if __name__ == '__main__':
    make_example_command(False)(['--help'])



from collections import OrderedDict
from typing import (
    Callable,
    Dict,
    Iterable,
    List,
    Optional,
    Sequence,
    Tuple,
    Type,
    Union,
    cast,
)

import click

from ._option_groups import OptionGroup, get_option_group_of


Subcommands = Union[Iterable[click.Command], Dict[str, click.Command]]


class GroupSection:
    """
    A section of commands inside a ``cloup.Group``. Sections are not
    (multi)commands, they simply allow to organize cloup.Group subcommands
    in many different help sections.
    """

    def __init__(self, title: str,
                 commands: Subcommands = (),
                 sorted: bool = False):  # noqa
        """
        :param title:
        :param commands: sequence of commands or dictionary {name: command}
        :param sorted:
            if True, ``list_commands()`` returns the commands in lexicographic order
        """
        self.title = title
        self.sorted = sorted  # type: ignore
        self.commands = OrderedDict()  # type: OrderedDict[str, click.Command]
        if isinstance(commands, Sequence):
            self.commands = OrderedDict()
            for cmd in commands:
                self.add_command(cmd)
        elif isinstance(commands, dict):
            self.commands = OrderedDict(commands)
        else:
            raise TypeError('commands must be a list of commands or a dict {name: command}')

    @classmethod
    def sorted(cls, title: str, commands: Subcommands = ()) -> 'GroupSection':
        return cls(title, commands, sorted=True)

    def add_command(self, cmd: click.Command, name: Optional[str] = None):
        name = name or cmd.name
        if not name:
            raise TypeError('missing command name')
        if name in self.commands:
            raise Exception('command "{}" already exists'.format(name))
        self.commands[name] = cmd

    def list_commands(self) -> List[Tuple[str, click.Command]]:
        command_list = [(name, cmd) for name, cmd in self.commands.items() if not cmd.hidden]
        if self.sorted:
            command_list.sort()
        return command_list

    def __len__(self) -> int:
        return len(self.commands)

    def __repr__(self) -> str:
        return 'GroupSection({}, sorted={})'.format(self.title, self.sorted)


class Command(click.Command):
    """ A ``click.Command`` supporting option groups. """

    def __init__(self, name, context_settings=None, callback=None, params=None, help=None,
                 epilog=None, short_help=None, options_metavar="[OPTIONS]", add_help_option=True,
                 hidden=False, deprecated=False, align_option_groups=True,
                 **kwargs):
        super().__init__(
            name=name, context_settings=context_settings, callback=callback, params=params,
            help=help, epilog=epilog, short_help=short_help, options_metavar=options_metavar,
            add_help_option=add_help_option, hidden=hidden,
            deprecated=deprecated, **kwargs)

        options_by_group = OrderedDict()
        for param in self.params:
            if isinstance(param, click.Argument):
                continue
            options_by_group.setdefault(get_option_group_of(param), []).append(param)

        self.ungrouped_options = options_by_group.pop(None, default=[])
        self.option_groups = list(options_by_group.keys())
        for group, options in options_by_group.items():
            group.options = options
        self.align_option_groups = align_option_groups

    def get_ungrouped_options(self, ctx: click.Context) -> Sequence[click.Option]:
        help_option = self.get_help_option(ctx)
        if help_option is not None:
            return self.ungrouped_options + [help_option]
        else:
            return self.ungrouped_options

    def format_option_group(self, ctx: click.Context,
                            formatter: click.HelpFormatter,
                            option_group: OptionGroup,
                            help_records: Optional[Sequence] = None):
        if help_records is None:
            help_records = option_group.get_help_records(ctx)
        if not help_records:
            return
        with formatter.section(option_group.name):
            if option_group.help:
                formatter.write_text(option_group.help)
            formatter.write_dl(help_records)

    def format_options(self, ctx: click.Context,
                       formatter: click.HelpFormatter,
                       max_option_width: int = 30):
        records_by_group = OrderedDict()  # OrderedDict for python 3.5
        for group in self.option_groups:
            records_by_group[group] = group.get_help_records(ctx)
        ungrouped_options = self.get_ungrouped_options(ctx)
        if ungrouped_options:
            default_group = OptionGroup('Other options' if records_by_group else 'Options')
            default_group.options = ungrouped_options
            records_by_group[default_group] = default_group.get_help_records(ctx)

        if self.align_option_groups:
            option_name_width = min(
                max_option_width,
                max(len(rec[0])
                    for records in records_by_group.values()
                    for rec in records)
            )
            # This is a hacky way to have aligned options groups.
            # Pad the first column of the first entry of each group to reach option_name_width
            for records in records_by_group.values():
                first = records[0]
                pad_width = option_name_width - len(first[0])
                if pad_width <= 0:
                    continue
                records[0] = (first[0] + ' ' * pad_width, first[1])

        for group, records in records_by_group.items():
            self.format_option_group(ctx, formatter, group, help_records=records)


class Group(click.Group):
    """
    A ``click.Group`` that supports subcommand help sections and returns
    and whose subcommands are, by default, of class ``cloup.Commands``.

    Subgroups can be specified in different ways:

    #. just pass a list of GroupSection objects to the constructor in ``sections``
    #. use ``add_section`` to add a section
    #. use ``add_command(cmd, name, section, ...)``
    #. use ``group.command(name, section, ...)``

    Commands not included in any user-defined section are added to the
    "default section", whose title is "Commands" or "Other commands" depending
    on whether it is the only section or not. The default section is the last
    shown section in the help and its commands are listed in lexicographic order.
    """

    def __init__(self, name: Optional[str] = None,
                 commands: Optional[Dict[str, click.Command]] = None,
                 sections: Iterable[GroupSection] = (),
                 align_sections: bool = True, **attrs):
        """
        :param name: name of the command
        :param commands: dict {name: command}; this command will be added to the default section.
        :param sections: a list of GroupSection objects
        :param align_sections: if True, the help column of all columns will be aligned;
            if False, each section will be formatted independently
        :param attrs:
        """
        super().__init__(name=name, commands=commands, **attrs)
        self.align_sections = align_sections
        self._default_section = GroupSection('__DEFAULT', commands=commands or [])
        self._user_sections = []  # type: List[GroupSection]
        self._section_set = set([self._default_section])
        for section in sections:
            self.add_section(section)

    def command(self, name: Optional[str] = None,  # type: ignore
                section: Optional[GroupSection] = None,
                cls: Type[click.Command] = Command, **attrs) -> Callable[[Callable], click.Command]:
        """ Creates a new command and adds it to this group. """

        def decorator(f):
            cmd = click.command(name=name, cls=cls, **attrs)(f)
            self.add_command(cmd, section=section)
            return cmd

        return decorator

    def group(self, name: Optional[str] = None,  # type: ignore
              section: Optional[GroupSection] = None,
              cls: Optional[Type[click.Group]] = None,
              **attrs) -> Callable[[Callable], click.Group]:
        if cls is None:
            cls = Group

        def decorator(f):
            cmd = click.group(name=name, cls=cls, **attrs)(f)
            self.add_command(cmd, section=section)
            return cmd

        return decorator

    def _add_command_to_section(self, cmd, name=None, section=None):
        """ Adds a command to the section (if specified) or to the default section """
        name = name or cmd.name
        if section is None:
            section = self._default_section
        section.add_command(cmd, name)
        if section not in self._section_set:
            self._user_sections.append(section)
            self._section_set.add(section)

    def add_section(self, section: GroupSection):
        """ Adds a :class:`GroupSection` to this group. You can add the same
        section object a single time. """
        if section in self._section_set:
            raise ValueError('section {} was already added'.format(section))
        self._user_sections.append(section)
        self._section_set.add(section)
        for name, cmd in section.commands.items():
            super().add_command(cmd, name)

    def section(self, title: str, *commands: click.Command, **attrs) -> GroupSection:
        """ Creates a new :class:`GroupSection`, adds it to this group and returns it. """
        section = GroupSection(title, commands, **attrs)
        self.add_section(section)
        return section

    def add_command(self, cmd: click.Command,
                    name: Optional[str] = None,
                    section: Optional[GroupSection] = None):
        """
        Adds a new command. If ``section`` is None, the command is added to the default section.
        """
        super().add_command(cmd, name)
        self._add_command_to_section(cmd, name, section)

    def list_sections(self, ctx: click.Context,
                      include_default_section: bool = True) -> List[GroupSection]:
        """ Returns the list of all sections in the "correct order".
         if ``include_default_section=True`` and the default section is non-empty,
         it will be included at the end of the list. """
        section_list = list(self._user_sections)
        if include_default_section and len(self._default_section) > 0:
            default_section = GroupSection.sorted(
                title='Other commands' if len(self._user_sections) > 0 else 'Commands',
                commands=self._default_section.commands)
            section_list.append(default_section)
        return section_list

    def format_commands(self, ctx: click.Context, formatter: click.HelpFormatter):
        section_list = self.list_sections(ctx)
        command_name_col_width = None
        if self.align_sections:
            command_name_col_width = max(len(name)
                                         for section in section_list
                                         for name in section.commands)
        for section in section_list:
            self.format_section(ctx, formatter, section, command_name_col_width)

    def format_section(self, ctx: click.Context,
                       formatter: click.HelpFormatter,
                       section: GroupSection,
                       command_col_width: Optional[int] = None):
        commands = section.list_commands()
        if not commands:
            return

        if command_col_width is None:
            command_col_width = max(len(cmd_name) for cmd_name, _ in commands)

        limit = formatter.width - 6 - command_col_width  # type: ignore
        rows = []
        for name, cmd in commands:
            short_help = cmd.get_short_help_str(limit)
            padded_name = name + ' ' * (command_col_width - len(name))
            rows.append((padded_name, short_help))

        if rows:
            with formatter.section(section.title):
                formatter.write_dl(rows)


def group(name: Optional[str] = None,
          cls: Type[Group] = Group, **attrs) -> Callable[[Callable], Group]:
    """ Creates a new ``Group`` (by default). """
    if not issubclass(cls, Group):
        raise TypeError('cls must be a subclass of cloup.Group')
    return cast(Group, click.group(name=name, cls=cls, **attrs))


def command(name: Optional[str] = None,
            cls: Type[Command] = Command, **attrs) -> Callable[[Callable], Command]:
    """ Creates a new ``cloup.Command`` (by default). """
    if not issubclass(cls, Command):
        raise TypeError('cls must be a subclass of cloup.Command')
    return cast(Command, click.command(name, cls=cls, **attrs))



"""Top-level package for cloup."""

__author__ = """Gianluca Gippetto"""
__email__ = 'gianluca.gippetto@gmail.com'
__version__ = '0.4.2'

# flake8: noqa F401
from ._option_groups import (
    GroupedOption, OptionGroup, option, option_group,
)
from ._commands import (
    Command, Group, GroupSection, command, group,
)




#!/usr/bin/env python
import click
import pytest
from click.testing import CliRunner

import cloup


def noop(*args, **kwargs):
    pass


@pytest.mark.parametrize('align_option_groups', [True, False], ids=['aligned', 'non-aligned'])
def test_example_command_help(align_option_groups, get_example_command):
    cmd = get_example_command(align_option_groups)
    runner = CliRunner()
    result = runner.invoke(cmd, args=('--help',))
    assert result.exit_code == 0
    assert result.output.strip() == cmd.expected_help


@pytest.mark.parametrize('align_sections', [True, False], ids=['aligned', 'non-aligned'])
def test_example_group_help(align_sections, get_example_group):
    grp = get_example_group(align_sections)
    result = CliRunner().invoke(grp, args=('--help',))
    if result.exception:
        raise result.exception
    assert result.exit_code == 0
    assert result.output.strip() == grp.expected_help


@pytest.mark.parametrize('method_name, cls, section', [
    ('command', click.Command, None),
    ('command', cloup.Command, cloup.GroupSection('A')),
    ('group', click.Group, cloup.GroupSection('A')),
    ('group', cloup.Group, None),
])
def test_Group_command_decorator(method_name, cls, section):
    grp = cloup.Group('ciao')
    decorator = getattr(grp, method_name)
    section = cloup.GroupSection('1')
    cmd = decorator('cmd', section=section, cls=cls, help='Help')(noop)
    assert cmd.__class__ is cls
    assert grp.commands['cmd'] is cmd
    assert cmd.help == 'Help'
    if section:
        assert grp._user_sections[0].commands['cmd'] is cmd
    else:
        assert grp._default_section.commands['cmd'] is cmd


def test_option_group_decorator_raises_if_group_is_passed_to_contained_option():
    func = cloup.option_group(
        'a group', cloup.option('--opt', group=cloup.OptionGroup('another group')))
    with pytest.raises(ValueError):
        func(noop)


def test_option_group_decorator_raises_for_no_options():
    with pytest.raises(ValueError):
        cloup.option_group('grp')



# flake8: noqa E128
import cloup
from cloup import GroupSection


def make_example_group(align_sections):
    def f(**kwargs):
        print(**kwargs)

    git_clone = cloup.command(
        'clone', help='Clone a repository into a new directory')(f)
    git_hidden1 = cloup.command(
        'hidden1', hidden=True)(f)
    git_init = cloup.command(
        'init', help='Create an empty Git repository or reinitialize an existing one')(f)

    git_rm = cloup.command(
        'rm', help='Remove files from the working tree and from the index')(f)
    git_sparse_checkout = cloup.command(
        'sparse-checkout', help='Initialize and modify the sparse-checkout')(f)
    git_mv = cloup.command(
        'mv', help='Move or rename a file, a directory, or a symlink')(f)

    @cloup.group(
        'git', align_sections=align_sections, context_settings={'terminal_width': 80})
    def git():
        return 0

    # We'll add commands/sections in all possible ways
    first_section = git.section(
        'Start a working area (see also: git help tutorial)', git_init, git_hidden1)
    first_section.add_command(git_clone)

    git.add_section(GroupSection(
        'Work on the current change (see also: git help everyday)',
        [git_rm, git_sparse_checkout, git_mv],
        sorted=True
    ))

    git.add_command(cloup.command('fake-3', hidden=True)(f))
    git.add_command(cloup.command('fake-2', help='Fake command #2')(f))
    git.add_command(cloup.command('fake-1', help='Fake command #1')(f))

    git.expected_help = (EXPECTED_ALIGNED_HELP if align_sections
                         else EXPECTED_NON_ALIGNED_HELP)
    return git


EXPECTED_ALIGNED_HELP = """
Usage: git [OPTIONS] COMMAND [ARGS]...

Options:
  --help  Show this message and exit.

Start a working area (see also: git help tutorial):
  init             Create an empty Git repository or reinitialize an existing...
  clone            Clone a repository into a new directory

Work on the current change (see also: git help everyday):
  mv               Move or rename a file, a directory, or a symlink
  rm               Remove files from the working tree and from the index
  sparse-checkout  Initialize and modify the sparse-checkout

Other commands:
  fake-1           Fake command #1
  fake-2           Fake command #2
""".strip()

EXPECTED_NON_ALIGNED_HELP = """
Usage: git [OPTIONS] COMMAND [ARGS]...

Options:
  --help  Show this message and exit.

Start a working area (see also: git help tutorial):
  init   Create an empty Git repository or reinitialize an existing one
  clone  Clone a repository into a new directory

Work on the current change (see also: git help everyday):
  mv               Move or rename a file, a directory, or a symlink
  rm               Remove files from the working tree and from the index
  sparse-checkout  Initialize and modify the sparse-checkout

Other commands:
  fake-1  Fake command #1
  fake-2  Fake command #2
""".strip()


if __name__ == '__main__':
    make_example_group(True)(['--help'])



from typing import Callable, Optional, Sequence, Type, overload

import click

#: A decorator that registers one or multiple click Options to the decorated function
OptionDecorator = Callable[[Callable], Callable]


class OptionGroup:
    def __init__(self, name: str, help: Optional[str] = None):
        if not name:
            raise ValueError('name is a mandatory argument')
        self.name = name
        self.help = help
        self.options = []  # type: Sequence[click.Option]

    def get_help_records(self, ctx: click.Context):
        return [opt.get_help_record(ctx) for opt in self if not opt.hidden]

    def option(self, *param_decls, **attrs):
        return option(*param_decls, group=self, **attrs)

    def __iter__(self):
        return iter(self.options)

    def __getitem__(self, i: int) -> click.Option:
        return self.options[i]

    def __len__(self) -> int:
        return len(self.options)

    def __repr__(self) -> str:
        return 'OptionGroup({!r}, help={!r}, options={})'.format(
            self.name, self.help, self.options)

    def __str__(self) -> str:
        return 'OptionGroup({!r}, help={!r}, options={})'.format(
            self.name, self.help, [opt.name for opt in self.options])


class GroupedOption(click.Option):
    """ A click.Option with an extra field ``group`` of type OptionGroup """

    def __init__(self, *args, group: Optional[OptionGroup] = None, **attrs):
        super().__init__(*args, **attrs)
        self.group = group


def has_option_group(param) -> bool:
    return hasattr(param, 'group') and param.group is not None


def get_option_group_of(param, default=None):
    return param.group if has_option_group(param) else default


def option(
    *param_decls,
    group: Optional[OptionGroup] = None,
    cls: Type[click.Option] = GroupedOption,
    **attrs
) -> OptionDecorator:
    def decorator(f):
        func = click.option(*param_decls, cls=cls, **attrs)(f)
        new_option = func.__click_params__[-1]
        new_option.group = group
        return func

    return decorator


@overload
def option_group(name: str, help: str, *options) -> OptionDecorator:
    ...


@overload
def option_group(name: str, *options, help: Optional[str] = None) -> OptionDecorator:
    ...


def option_group(name: str, *args, **kwargs) -> OptionDecorator:
    """
    Attaches an option group to the command. This decorator is overloaded with
    two signatures::

        @option_group(name: str, *options, help: Optional[str] = None)
        @option_group(name: str, help: str, *options)

    In other words, if the second position argument is a string, it is interpreted
    as the "help" argument. Otherwise, it is interpreted as the first option;
    in this case, you can still pass the help as keyword argument.
    """
    if args and isinstance(args[0], str):
        return _option_group(name, options=args[1:], help=args[0])
    else:
        return _option_group(name, options=args, **kwargs)


def _option_group(
    name: str, options: Sequence[Callable], help: Optional[str] = None
) -> OptionDecorator:
    if not options:
        raise ValueError('you must provide at least one option')

    opt_group = OptionGroup(name, help=help)

    def decorator(f):
        for opt_decorator in reversed(options):
            # Note: the assignment is just a precaution, as both click.option
            # and @cloup.option currently return the same f
            f = opt_decorator(f)
            new_option = f.__click_params__[-1]
            curr_group = get_option_group_of(new_option)
            if curr_group is not None:
                raise ValueError(
                    'option {} was first assigned to {} and then passed '
                    'as argument to @option_group({!r}, ...)'
                    .format(new_option.opts, curr_group, name))
            new_option.group = opt_group
        return f

    return decorator



#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('CHANGELOG.rst') as history_file:
    history = history_file.read()

setup(
    # Name, author, description and other info
    name='cloup',
    version='0.4.2',
    author="Gianluca Gippetto",
    author_email='gianluca.gippetto@gmail.com',
    description="Option groups and subcommand help sections for pallets/click",
    long_description=readme + '\n\n' + history,
    long_description_content_type='text/x-rst',
    url='https://github.com/janLuke/cloup',
    license="MIT license",
    keywords='cloup click option',
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    packages=find_packages(include=['cloup', 'cloup.*']),
    zip_safe=False,
    include_package_data=True,
    python_requires='>=3.5',
    install_requires=[
        'click>=7.0,<9.0'
    ],
)
